import requests
import os

from typing import List

API_BASE_URL = os.environ.get("API_BASE_URL", "https://data-idm.unavco.org/user/profile")


def get_user(access_token: str):
    r = requests.get(
        f"{API_BASE_URL}/api/user/",
        headers={"authorization": f"Bearer {access_token}"},
    )
    if r.status_code == 200:
        return r.json()

    raise RuntimeError(r.json()["detail"])


def lookup_anon(access_token: str, ids: List[str], emails: List[str]):
    r = requests.post(
        f"{API_BASE_URL}/api/user/anon/",
        headers={"authorization": f"Bearer {access_token}"},
        json={"ids": ids, "emails": emails},
    )

    if r.status_code == 200:
        return r.json()

    raise RuntimeError(r.json()["detail"])





import datetime as dt
import logging

from dataclasses import dataclass
from typing import Dict, Optional, Union
import jwt

import requests

logger = logging.getLogger(__name__)


class AuthFlowError(Exception):
    """Generic authentication flow error"""


class UnauthorizedError(AuthFlowError):
    pass


class NoTokensError(AuthFlowError):
    pass


class NoIdTokenError(NoTokensError):
    pass


class NoRefreshTokenError(NoTokensError):
    pass


class InvalidRefreshTokenError(AuthFlowError):
    pass


@dataclass
class TokensResponse:
    access_token: str
    refresh_token: Optional[str]
    id_token: Optional[str]
    token_type: str
    expires_in: int
    scope: Optional[str]


@dataclass
class UnvalidatedTokens:
    access_token: str
    id_token: Optional[str]
    refresh_token: Optional[str]
    scope: str

    @classmethod
    def from_dict(cls, o: Dict):
        return cls(
            access_token=o.get("access_token"),
            id_token=o.get("id_token"),
            refresh_token=o.get("refresh_token"),
            scope=o.get("scope"),
        )


@dataclass
class ValidTokens(UnvalidatedTokens):
    expires_at: int
    issued_at: int

    @classmethod
    def from_unvalidated(cls, creds: UnvalidatedTokens, body: Dict):
        return cls(
            access_token=creds.access_token,
            id_token=creds.id_token,
            refresh_token=creds.refresh_token,
            scope=creds.scope,
            expires_at=body.get("exp"),
            issued_at=body.get("iat"),
        )


class AuthFlow:
    @property
    def access_token(self):
        return self.tokens.access_token

    @property
    def access_token_body(self):
        if self._access_token_body is None:
            raise NoTokensError
        return self._access_token_body

    @property
    def tokens(self):
        if not self._tokens:
            raise NoTokensError

        return self._tokens

    @property
    def expires_at(self):
        return dt.datetime.fromtimestamp(self.tokens.expires_at, dt.timezone.utc)

    @property
    def id_token(self):
        idt = self.tokens.id_token
        if not idt:
            raise NoIdTokenError

        return idt

    @property
    def id_token_body(self):
        if self._id_token_body is None:
            raise NoIdTokenError
        return self._id_token_body

    @property
    def issued_at(self):
        return dt.datetime.fromtimestamp(self.tokens.issued_at, dt.timezone.utc)

    @property
    def refresh_token(self):
        rt = self.tokens.refresh_token
        if not rt:
            raise NoRefreshTokenError

        return rt

    @property
    def refreshable(self):
        try:
            return self.tokens.refresh_token is not None
        except NoTokensError:
            return False

    @property
    def ttl(self):
        return self.expires_at - dt.datetime.now(dt.timezone.utc)

    def __init__(self, domain: str, audience: str, client_id: str, scope: str) -> None:
        self.auth0_domain = domain
        self.auth0_audience = audience
        self.auth0_client_id = client_id
        self.token_scope = scope

        self._tokens: Optional[ValidTokens] = None
        self._access_token_body: Optional[Dict[str, Union[str, int]]] = None
        self._id_token_body: Optional[Dict[str, Union[str, int]]] = None

    def load_tokens(self):
        raise NotImplementedError

    def save_tokens(self, creds: ValidTokens):
        logger.warning(
            "Token is not actually persisted. Override this method to handle persistence"
        )

    def refresh(self, scope: Optional[str] = None, persist=True):
        scope = scope or self.token_scope

        r = requests.post(
            f"https://{self.auth0_domain}/oauth/token",
            headers={"content-type": "application/x-www-form-urlencoded"},
            data={
                "grant_type": "refresh_token",
                "client_id": self.auth0_client_id,
                "refresh_token": self.refresh_token,
                "scopes": scope,
            },
        )

        if r.status_code == 200:
            if not persist:
                return self

            self.validate_and_save_tokens(r.json())
            logger.debug(f"Refreshed tokens: {self._tokens}")
            return self

        raise InvalidRefreshTokenError

    def validate_and_save_tokens(self, unvalidated_creds: Dict):
        self.validate_tokens(unvalidated_creds)
        try:
            self.save_tokens(self._tokens)
        except Exception as e:
            logger.error("Error while persisting tokens", exc_info=e)

        return self

    def validate_tokens(self, unvalidated_creds: Dict):
        creds = UnvalidatedTokens.from_dict(unvalidated_creds)
        idt_body: Optional[Dict] = None

        # TODO: validate tokens
        logger.warning("NO TOKEN VALIDATION IMPLEMENTED")
        try:
            at_body = jwt.decode(
                creds.access_token,
                options={"verify_signature": False},
            )

            if creds.id_token:
                idt_body = jwt.decode(
                    creds.id_token,
                    options={"verify_signature": False},
                )
        except Exception as e:
            logger.error("Invalid tokens", exc_info=e)
            raise

        self._access_token_body = at_body
        self._id_token_body = idt_body
        self._tokens = ValidTokens.from_unvalidated(
            creds=creds,
            body=at_body,
        )
        self.token_scope = self._tokens.scope

        return self



from dataclasses import dataclass
import logging
import requests

from earthscope_sdk.auth.auth_flow import (
    AuthFlow,
    AuthFlowError,
    UnauthorizedError,
)

logger = logging.getLogger(__name__)


class ClientCredentialsFlowError(AuthFlowError):
    pass


@dataclass
class ClientCredentials:
    client_id: str
    client_secret: str


@dataclass
class GetTokensErrorResponse:
    error: str
    error_description: str


class ClientCredentialsFlow(AuthFlow):
    def __init__(
        self, domain: str, audience: str, client_credentials: ClientCredentials
    ) -> None:
        if not client_credentials or not client_credentials.client_secret:
            raise ValueError("Client secret missing")

        super().__init__(
            domain=domain,
            audience=audience,
            client_id=client_credentials.client_id,
            scope="",
        )

        # Flow management vars
        self._secret = client_credentials.client_secret

    def request_tokens(self):
        r = requests.post(
            f"https://{self.auth0_domain}/oauth/token",
            headers={"content-type": "application/x-www-form-urlencoded"},
            data={
                "grant_type": "client_credentials",
                "client_id": self.auth0_client_id,
                "client_secret": self._secret,
                "audience": self.auth0_audience,
            },
        )
        if r.status_code == 200:
            self.validate_and_save_tokens(r.json())
            logger.debug(f"Got tokens: {self.tokens}")
            return self

        if r.status_code == 401:
            err = GetTokensErrorResponse(**r.json())
            if err.error == "access_denied":
                if err.error_description == "Unauthorized":
                    raise UnauthorizedError

        raise ClientCredentialsFlowError



import logging
import requests

from dataclasses import dataclass
from enum import Enum
from time import sleep
from typing import Optional

from earthscope_sdk.auth.auth_flow import AuthFlow


logger = logging.getLogger(__name__)


class PollingErrorType(str, Enum):
    AUTHORIZATION_PENDING = "authorization_pending"
    SLOW_DOWN = "slow_down"
    EXPIRED_TOKEN = "expired_token"
    ACCESS_DENIED = "access_denied"


class RequestDeviceTokensError(RuntimeError):
    pass


class PollingError(ValueError):
    pass


class PollingExpiredError(PollingError):
    pass


class PollingAccessDeniedError(PollingError):
    pass


@dataclass
class GetDeviceCodeResponse:
    device_code: str
    user_code: str
    verification_uri: str
    verification_uri_complete: str
    expires_in: int
    interval: int


@dataclass
class PollingErrorResponse:
    error: PollingErrorType
    error_description: str


class DeviceCodeFlow(AuthFlow):
    @property
    def polling(self):
        return self._is_polling

    @property
    def started(self):
        return self._device_codes is not None

    def __init__(self, domain: str, audience: str, client_id: str, scope: str) -> None:
        super().__init__(
            domain=domain,
            audience=audience,
            client_id=client_id,
            scope=scope,
        )

        # Flow management vars
        self._is_polling = False
        self._device_codes: Optional[GetDeviceCodeResponse] = None

    def do_flow(self):
        self.request_device_code()
        self.prompt_user()
        self.poll()
        return self.tokens

    def poll(self):
        if not self._device_codes:
            raise PollingError("Cannot poll without initial device code response")

        if self._is_polling:
            raise PollingError("Attempted to double poll")

        self._is_polling = True
        try:
            while True:
                sleep(self._device_codes.interval)

                r = requests.post(
                    f"https://{self.auth0_domain}/oauth/token",
                    headers={"content-type": "application/x-www-form-urlencoded"},
                    data={
                        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                        "device_code": self._device_codes.device_code,
                        "client_id": self.auth0_client_id,
                    },
                )
                if r.status_code == 200:
                    tokens = self.validate_and_save_tokens(r.json())
                    logger.debug(f"Got tokens: {tokens}")
                    return self

                poll_err = PollingErrorResponse(**r.json())
                if poll_err.error in [
                    PollingErrorType.AUTHORIZATION_PENDING,
                    PollingErrorType.SLOW_DOWN,
                ]:
                    continue

                if poll_err.error == PollingErrorType.EXPIRED_TOKEN:
                    raise PollingExpiredError

                if poll_err.error == PollingErrorType.ACCESS_DENIED:
                    raise PollingAccessDeniedError

                if poll_err:
                    raise PollingError(f"Unknown polling error: {poll_err}")
        finally:
            self._is_polling = False

    def prompt_user(self):
        raise NotImplementedError(
            "This class cannot prompt a user. Create a subclass that implements this method."
        )

    def request_device_code(self):
        r = requests.post(
            f"https://{self.auth0_domain}/oauth/device/code",
            headers={"content-type": "application/x-www-form-urlencoded"},
            data={
                "client_id": self.auth0_client_id,
                "scope": self.token_scope,
                "audience": self.auth0_audience,
            },
        )
        if r.status_code == 200:
            self._device_codes = GetDeviceCodeResponse(**r.json())
            logger.debug(f"Got device code response: {self._device_codes}")
            return self

        raise RequestDeviceTokensError(f"Failed to get a device code: {r.text}")
